import numpy as np
import scipy as sp
from matplotlib import pyplot as plt


Lg = 20e-6  # Grid inductance
Lf = 50e-6
Rf = 1e-3

## Current control loop
# BW is determined by the Kp and the crossover frequency
fc_i = 5e3
# tr_i = 0.35 / fc_i
# wc_i = 2 * np.pi * fc_i
# kp_i = wc_i * Lf

# Resonant frequency is 55Hz, so that we can easily cover for both 50 and 60Hz, with a reasonably wide resonance
f0 = 55
# w0 = 2 * np.pi * f0


def arg_approx(kp, w_cross, ksi_r, w0, gain_max):
    """
    Calculate the argument for an array of max gain values.
    """
    wr = ksi_r * w0
    arg_vals = []
    for gain_val in gain_max:
        num1 = 2 * wr * gain_val * w_cross
        num0 = kp * (w0**2-w_cross**2)
        den1 = 2 * wr * w_cross
        den0 = w0**2-w_cross**2

        arg_val = 180 / np.pi * (np.arctan2(num1, num0) - np.arctan2(den1, den0))
        arg_vals.append(arg_val)
    return arg_vals


def draw_PR_bode(kp, ki, ksi, w0):
    num = [kp, 2 * ksi * w0 * (kp + ki), kp * w0 ** 2]
    denum = [1, 2 * ksi * w0, w0 ** 2]

    lti = sp.signal.lti(num, denum)
    w = [2 * np.pi * f for f in np.logspace(1, 4, 1000)]
    w, mag, phase = sp.signal.bode(lti, w=w)
    f = np.divide(w, 2 * np.pi)

    figure, axis = plt.subplots(2, 1)
    axis[0].semilogx(f, mag)  # Bode magnitude plot
    axis[0].set_title("Mag")
    axis[1].semilogx(f, phase)  # Bode phase plot
    axis[1].set_title("Phase")
    axis[1].axvline(x=fc_i, color='r')
    plt.show()


def draw_notch_bode(ksi, f0, fp):
    w0 = 2 * np.pi * f0
    wp = 2 * np.pi * fp
    num = [1, 0, w0 ** 2]
    denum = [1, 2 * ksi * wp, wp ** 2]

    lti = sp.signal.lti(num, denum)
    w = [2 * np.pi * f for f in np.logspace(1, 4, 1000)]
    w, mag, phase = sp.signal.bode(lti, w=w)
    f = np.divide(w, 2 * np.pi)

    figure, axis = plt.subplots(2, 1)
    axis[0].semilogx(f, mag)  # Bode magnitude plot
    axis[0].set_title("Mag")
    axis[1].semilogx(f, phase)  # Bode phase plot
    axis[1].set_title("Phase")
    axis[0].axvline(x=50, color='r')
    axis[0].axvline(x=60, color='r')
    plt.show()


# The goal is to find the largest Ki that doesn't threaten the stability

def calc_current_ctrl_params(l_filter, f_cross, phase_lag, f_res=55, ksi_r=0.1):
    tr_i = 0.35 / f_cross
    w0 = 2 * np.pi * f_res
    w_cross = 2 * np.pi * f_cross
    kp = w_cross * l_filter

    gain_max_vec = np.arange(10, 1000, step=10)
    arg_vals = arg_approx(kp, w_cross, ksi_r, w0, gain_max_vec)
    # This is the maximum phase lag that the resonant controller should add
    # phase_lag_max = 30
    # Maximum gain of the controller that will not increase the phase lag over the maximum values
    gain_max = np.interp(phase_lag, np.multiply(arg_vals, -1), gain_max_vec)
    # Maximum gain happens at resonant frequency and is equal to ki + kp
    ki = gain_max - kp

    return kp, ki


def calc_dc_cap(p_n, vg_n, vdc_n, delta_u=10):
    """
    In a single-phase system, power flow oscillates at twice the grid frequency.
    Capacitor is designed to limit the voltage oscillations at this frequency.
    Total energy in one half-period of power oscillation is U_rms*I_rms/w_grid.
    This energy is pumped into the capacitor during a quarter of the grid period.

    C/2 * [(U + dU)**2 - U**2] = E
    """
    i_n = p_n / vg_n
    delta_e = vg_n * i_n / (2 * np.pi * 50)
    cap_val = 2 * delta_e / ((vdc_n + delta_u)**2 - vdc_n**2)
    return cap_val


def calc_volt_ctrl_params(vdc_n, vg_n, cap_val, f_cross):
    w_cross = 2 * np.pi * f_cross
    kp = cap_val * w_cross * vdc_n / vg_n
    ki = kp * w_cross / 10
    return kp, ki


if __name__ == "__main__":
    kp_i, ki_i = calc_current_ctrl_params(Lf, 5e3, 30)
    draw_PR_bode(kp_i, ki_i, 0.1, 2 * np.pi * 55)
    # draw_notch_bode(10, 55, 55)
