import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
import typhoon.test.reporting.figures as fig
from pandas import Timedelta as td
from typhoon.test.harmonic import frequency_content
import logging
import pytest
import os
from pytest import assume
import numpy as np


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located

tests_dir = os.path.dirname(FILE_DIR_PATH)

logger = logging.getLogger(__name__)

MODEL_NAME = "blinky.tse"

# .tse file paths
MODEL_PATH = os.path.join(os.path.dirname(tests_dir), 'models', 'blinky', 'blinky.tse')

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)

GPIO31_BLINKY_DEVICES = [
    "LAUNCHXL-F28379D", "TMDSCNCD28379D", "TMDSCNCD28335"
]


@pytest.fixture(scope="module")
def setup_function(get_hw_setup):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Generic GPIO DO component is used to toggle HIL DI 1.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type='component')
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    subsys_handle = mdl.get_item('blinky', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    # Configure generic gpio comp regarding interface and platform setup
    generic_gpio_comp = mdl.get_item("GPIO - HIL DI1", parent=subsys_handle)
    mdl.set_property_value(
        mdl.prop(generic_gpio_comp, "interface_type"), interface_board
    )
    mdl.set_property_value(
        mdl.prop(generic_gpio_comp, "controller_index"), "1"
    )
    mdl.set_property_value(
        mdl.prop(generic_gpio_comp, "hil_di_nr"), "1"
    )

    if platform not in GPIO31_BLINKY_DEVICES:
        blinky_gpio_comp = mdl.get_item("GPIO DO1", parent=subsys_handle)
        mdl.set_property_value(
            mdl.prop(blinky_gpio_comp, "gpio"), "1"
        )

    # Generate and flash the code
    button_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(button_action_prop, True)
    mdl.set_property_value(button_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    hw_settings = mdl.detect_hw_settings()
    logger.info("{} {} device is used".format(hw_settings[0], hw_settings[2]))
    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)

    yield
    hil.stop_simulation()
    mdl.close_model()


def test_blinky(setup_function):
    """
    Test checks whether the HIL digital input 1 is receiving a blinking signal with 2.5Hz frequency.
    """

    capture_duration = 15
    ti_signal_name_1 = "HIL0 digital input 1"
    ti_signals = [ti_signal_name_1]
    capture.start_capture(duration=capture_duration, signals=ti_signals, executeAt=500e-6, rate=100e3)

    # start simulation, wait for capture to finish, stop simulation
    hil.start_simulation()

    cap_data = capture.get_capture_results(wait_capture=True)

    hil.stop_simulation()

    expected_freq = 2.5

    freq_cont = frequency_content(signal=cap_data[ti_signal_name_1], max_frequency=int(expected_freq*2))
    target_harmonic_amp = _calc_freq_amp(freq_cont, expected_freq, 0.2)
    logger.info(f"Harmonic amp at {expected_freq}: {target_harmonic_amp}")

    fig.attach_figure([cap_data[ti_signal_name_1][:td(2, 's')]], f'{expected_freq}Hz')

    with assume:
        assert target_harmonic_amp == pytest.approx(0.5*4/np.pi, rel=0.05), "Failure of harmonic content at targeted " \
                                                                            f"frequency ({expected_freq}Hz)!"


def _calc_freq_amp(freq_content, target_freq, rel_tol):
    delta_f = rel_tol * target_freq
    delta_f_step = delta_f * 0.02

    amp_val = 0
    previous_amp_vals = []
    for f_val in np.arange(target_freq-delta_f, target_freq+delta_f, delta_f_step):
        try:
            new_amp_val = abs(freq_content(f_val, tol=delta_f_step/2))
            if new_amp_val not in previous_amp_vals:
                amp_val = np.sqrt(amp_val ** 2 + new_amp_val ** 2)
        except ValueError:
            pass

    return amp_val
