import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
from typhoon.test.signals import assert_is_step
import typhoon.test.ranges as ranges
import logging
import pytest
import os


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located

tests_dir = os.path.dirname(FILE_DIR_PATH)

logger = logging.getLogger(__name__)

MODEL_NAME = "closed_loop_dab.tse"

# .tse file paths
MODEL_PATH = os.path.join(
    os.path.dirname(tests_dir), 'models', 'closed_loop_dab', 'closed_loop_dab.tse'
)

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)


@pytest.fixture(scope="module")
def setup_function(request, get_hw_setup):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Generic ADC component is used to sample the current.
    Analog pin availability and signal scaling and offset may vary regarding selected platform and interface board.
    Generic ePWM component is used to configure 4 synchronized modules with variable phase shift.
    HIL digital inputs for pwm signals may vary regarding selected platform and interface board.
    Availability of 4 consecutive ePWM modules may vary on selected platform and interface board.
    Generic GPIO DI component is used to sample HIL DI1 for PI controller integrator reset.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device

    i_max_sample = 150
    div = ti_hw_params.scaling
    scaling = i_max_sample / div
    offset = ti_hw_params.offset

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    hw_settings = mdl.detect_hw_settings()
    product = hw_settings[0]
    configuration = hw_settings[2]
    logger.info("{} c{} device is used".format(product, configuration))
    pytest.check_hw_requirements(request, product, configuration)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type=ITEM_COMPONENT)
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    subsys_handle = mdl.get_item('ti_exp', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    adc_comp = mdl.get_item("ADC (Generic)1", parent=subsys_handle)
    gpio_comp = mdl.get_item("GPIO - HIL DO1", parent=subsys_handle)
    epwm_comp = mdl.get_item("ePWM (Generic)1", parent=subsys_handle)

    for comp in [adc_comp, gpio_comp, epwm_comp]:
        # Just to trigger property_value_changed handler so the platform will be taken into account
        mdl.set_property_value(
            mdl.prop(comp, "interface_type"), "HIL DSP Interface"
        )
        mdl.set_property_value(
            mdl.prop(comp, "interface_type"), interface_board
        )
        mdl.set_property_value(
            mdl.prop(comp, "controller_index"), "1"
        )

    mdl.set_property_value(
        mdl.prop(gpio_comp, "hil_do_nr"), "1"
    )

    mdl.set_property_value(
        mdl.prop(epwm_comp, "nr_of_channels"), "4"
    )

    mdl.set_property_value(
        mdl.prop(adc_comp, "hil_ao_nr"), "1"
    )
    mdl.set_property_value(
        mdl.prop(adc_comp, "scaling"), scaling
    )
    mdl.set_property_value(
        mdl.prop(adc_comp, "offset"), offset
    )

    hil_digital_inputs_pwm = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr")
    )
    hil_digital_inputs_pwm_b = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr_b")
    )

    button_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(button_action_prop, True)
    mdl.set_property_value(button_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    # HIL DI number property od ePWM (Generic) displays only HIL pins that are mapped
    # to ePWM A channels. B channel connection is dependent on interface board type (see interface documentation)
    channel_a_digital_inputs = hil_digital_inputs_pwm.split("-")
    channel_b_digital_inputs = hil_digital_inputs_pwm_b.split("-")

    converter_handle = mdl.get_item("Dual Active Bridge1", parent=None)

    # Switches that are controlled by ePWM A channels
    switches_a = ["SA_1", "SA_3", "SB_1", "SB_3"]
    for switch, hil_di in zip(switches_a, channel_a_digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), hil_di
        )

    # Switches that are controlled by ePWM B channels
    switches_b = ["SA_2", "SA_4", "SB_2", "SB_4"]
    for switch, hil_di in zip(switches_b, channel_b_digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), hil_di
        )

    analog_dict = {
        "ao_channel": 1,
        "signal": "Iout",
        "scaling": scaling,
        "offset": offset,
        "enable_limit": "True",
        "lower_limit": ti_hw_params.ao_limits[0],
        "upper_limit": ti_hw_params.ao_limits[1]
    }
    output_settings = mdl.get_item("Output Settings1", parent=None)
    mdl.set_property_value(
        mdl.prop(output_settings, "analog_outputs"), [analog_dict]
    )

    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)

    serial_con = pytest.ti_serial()
    yield serial_con
    hil.stop_simulation()
    serial_con.close_connection()
    mdl.close_model()


@pytest.mark.parametrize("i_start, i_end", [(50, 100), (120, 70)])
def test_closed_loop_dab(setup_function, i_start, i_end):
    """
    Test checks whether the closed loop control of DAB converter is following
    the reference.
    """

    serial_con = setup_function
    capture_duration = 0.1
    step_timestamp = 0.2 * capture_duration
    capture_signals = ["Iout"]

    # start simulation, wait for capture to finish, stop simulation
    hil.start_simulation()

    # Set reference to 50
    serial_con.transmit(i_start, 1)

    hil.wait_msec(100)
    edge_type = 'rising' if i_end > i_start else 'falling'
    capture.start_capture(
        duration=capture_duration, signals=capture_signals, offset_absolute=step_timestamp,
        trigger_source="Iout", trigger_threshold=(i_end+i_start)/2, trigger_edge=edge_type,
        rate=500e3, timeout=10
    )

    # Set reference to 100
    serial_con.transmit(i_end, 1)

    cap_data = capture.get_capture_results(wait_capture=True)
    hil.stop_simulation()

    assert_is_step(
        signal=cap_data["Iout"], from_value=ranges.around(i_start, tol_p=0.05),
        to_value=ranges.around(i_end, tol_p=0.05), find_start=False, strictness=0.99,
        at_t=ranges.around(0, tol=1e-3)
    )
