import pytest
import params
import os
import sys


TESTS_DIR_PATH = os.path.dirname(__file__)  # path of the directory where this script is located

if TESTS_DIR_PATH not in sys.path:
    sys.path.append(TESTS_DIR_PATH)

from ti_example_test_utilities import TIPySerial, check_hw_requirements

pytest.ti_serial = TIPySerial
pytest.check_hw_requirements = check_hw_requirements


class InterfaceParams:
    def __init__(self, interface_name: str, ti_device: str, processor_type: str):
        self.interface_name = interface_name
        self.ti_device = ti_device
        self.ti_processor = processor_type
        self.scaling = 1.5
        self.offset = 1.5
        self.ao_limits = (0, 3)
        if self.ti_device in params.DEVICE_3V3:
            self.scaling = 1.65
            self.offset = 1.65
            self.ao_limits = (0, 3.33)
        elif self.ti_device in params.DEVICE_5V0:
            self.scaling = 5
            self.offset = 0
            self.ao_limits = (-5, 5)



def pytest_addoption(parser):
    parser.addoption(
        '--interface',
        action='store',
        default='HIL TI Launchpad Interface',
        help='Interface board definition'
    )
    parser.addoption(
        '--tidev',
        action='store',
        default='F28379D',
        help='Texas instruments device definition'
    )


@pytest.fixture(autouse=True, scope="module")
def get_interface_board(request):
    argument = str(request.config.getoption('--interface')).lower()
    if "grid" in argument:
        if any(char.isdigit() for char in argument):
            mcu_id = argument[len(argument) - 1]
            if mcu_id != '1':
                pytest.skip(
                    "--interface: MCU id is not valid. Slot 1 is supported for example tests.\n"
                    "See examples.pdf for instructions on how to run the tests properly."
                )
            else:
                interface = "HIL TI uGrid Launchpad Interface"
        else:
            interface = "HIL TI uGrid Launchpad Interface"

    elif "default" in argument or "launchpad" in argument:
        interface = "HIL TI Launchpad Interface"

    elif "180" in argument:
        interface = "HIL DSP 180 Interface"

    elif "100" in argument:
        interface = "HIL DSP Interface"

    else:
        pytest.skip(
            "Invalid --interface argument\n"
            "Available values:\n"
            "--interface=launchpad - HIL TI Launchpad Interface\n"
            "--interface=ugrid1 - HIL TI uGrid Launchpad Interface (slot 1)\n"
            "--interface=ugrid2 - HIL TI uGrid Launchpad Interface (slot 2) - not supported by example tests\n"
            "--interface=ugrid3 - HIL TI uGrid Launchpad Interface (slot 3) - not supported by example tests\n"
            "--interface=180 - HIL DSP 180 Interface\n"
            "--interface=100 - HIL DSP Interface\n"
            "See examples.pdf for instructions on how to run the tests properly."
        )

    test_file_name = os.path.basename(request.fspath)
    if interface not in params.TEST_SUPPORT_INTERFACE[test_file_name]:
        pytest.skip(
            f"Test is not supported by {interface}.\n"
            f"See examples.pdf for instructions on how to run the tests properly."
        )

    return interface


@pytest.fixture(autouse=True, scope="module")
def get_hw_setup(request, get_interface_board: str):
    device = str(request.config.getoption('--tidev')).upper()
    interface_board = get_interface_board
    supported_devices = params.INTERFACE_SUPPORT[
        interface_board
    ]
    if device not in supported_devices:
        supported_devices_print = '\n'.join(supported_devices)
        pytest.skip(
            f'Interface board and TI device are not compatible!\n'
            f'Available values of --tidev that are compatible with {interface_board}:\n'
            f'{supported_devices_print}\n'
            f'See examples.pdf for instructions on how to run the tests properly.'
        )

    if 'Launchpad' in interface_board and device in params.LAUNCHPAD_DEVICES:
        full_device_name = 'LAUNCHXL-' + device
    elif device in params.CONTROL_CARD_DEVICES:
        full_device_name = 'TMDSCNCD' + device.lstrip('F')
    else:
        pytest.skip(
            'Interface board and TI device are not compatible!\n'
            'See examples.pdf for instructions on how to run the tests properly.'
        )

    test_file_name = os.path.basename(request.fspath)
    if device not in params.TEST_SUPPORT_DEVICE[test_file_name]:
        pytest.skip(
            f"Test is not supported by {device} device.\n"
            f"See examples.pdf for instructions on how to run the tests properly."
        )

    return InterfaceParams(
        interface_board, full_device_name, device
    )


@pytest.fixture(scope='session', autouse=True)
def validate_ti_c2000_settings():
    ti_c2000_settings_path = os.path.join(
        params.LIB_PATH, "ti_c2000_settings.json"
    )
    if not os.path.exists(ti_c2000_settings_path):
        pytest.skip(
            "Settings file not found. Please write settings in TI C2000 Setup component in Schematic Editor.\n"
            "See examples.pdf for instructions on how to run the tests properly."
        )
