import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
from typhoon.test.signals import assert_is_constant
import typhoon.test.ranges as ranges
import logging
import pytest
import os
import allure
from contextlib import contextmanager
import params


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

logger = logging.getLogger(__name__)

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located
TESTS_DIR = os.path.dirname(FILE_DIR_PATH)

MODEL_NAME = "foc_pmsm.tse"

# .tse file paths
MODEL_PATH = os.path.join(
    os.path.dirname(TESTS_DIR), 'models', 'foc_pmsm', MODEL_NAME
)

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)


ENCODER_DIGITAL_OUTPUTS = [
    ['20', '21', '22'],     # TI Launchpad Interface
    ['1', '2', '4'],        # TI DSP 180 Interface
    ['1', '2', '3']         # TI DSP Interface
]

CURRENTS_ANALOG_OUTPUTS = [
    ['4', '5', '6'],        # TI Launchpad Interface
    ['7', '1', '9'],        # TI DSP 180 Interface
    ['4', '5', '10']        # TI DSP Interface
]


@pytest.fixture(scope="module")
def get_hil_outputs(get_interface_board: str) -> dict:
    """
    Fixture that provides dictionary with list of HIL analog and digital outputs connected to proper
    TI ADC and encoder peripheral channels.

    Returns:
        dict: analog: list of HIL AO numbers connected to proper ADC channels,
              digital: list of HIL DO numbers connected to eQEP A, B and Z channels respectively.
    """
    interface_board = get_interface_board
    if interface_board == params.INTERFACE_LAUNCHPAD:
        config_index = 0
    elif interface_board == params.INTERFACE_180:
        config_index = 1
    else:
        config_index = 2

    return {
        'analog': CURRENTS_ANALOG_OUTPUTS[config_index],
        'digital': ENCODER_DIGITAL_OUTPUTS[config_index]
    }


@contextmanager
def group_context(description):
    try:
        with allure.step(description):
            yield
    except AssertionError as e:
        with pytest.assume:
            raise e


@pytest.fixture(scope="module")
def setup(request, get_hw_setup, get_hil_outputs):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Generic ADC components is used to sample phase currents.
    Analog pin availability and signal scaling and offset may vary regarding selected platform and interface board.
    Generic ePWM component is used to configure 3 synchronized modules.
    HIL digital inputs for pwm signals may vary regarding selected platform and interface board.
    Availability of 3 consecutive ePWM modules may vary on selected platform and interface board.
    Generic GPIO DI components are used for PI controller integrator reset and control mode (speed/torque) - HIL DO 6&9
    eQEP component is used for position sensing, providing value of mechanical angle in every fast interrupt routine.
    Serial communication provides speed and torque references and sends calculated values to the PC.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device

    i_max_sample = 400
    div = ti_hw_params.scaling
    scaling = i_max_sample / div
    offset = ti_hw_params.offset

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    hw_settings = mdl.detect_hw_settings()
    product = hw_settings[0]
    configuration = hw_settings[2]
    logger.info("{} c{} device is used".format(product, configuration))
    pytest.check_hw_requirements(request, product, configuration)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type=ITEM_COMPONENT)
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    if platform == "TMDSCNCD28335":
        scheduling_interrupt, adc_trigger = ("ADC Interrupt", "ePWM")
    elif platform in ["LAUNCHXL-F280039C", "LAUNCHXL-F2800137"]:
        scheduling_interrupt, adc_trigger = ("ADCC Interrupt", "ePWM4")
    else:
        scheduling_interrupt, adc_trigger = ("ADCC Interrupt", "ePWM1")

    mdl.set_property_value(
        mdl.prop(setup_handle, "scheduling"), scheduling_interrupt
    )

    subsys_handle = mdl.get_item('foc', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    adc_comp1 = mdl.get_item("ADC (Generic)1", parent=subsys_handle)
    adc_comp2 = mdl.get_item("ADC (Generic)2", parent=subsys_handle)
    adc_comp3 = mdl.get_item("ADC (Generic)3", parent=subsys_handle)
    gpio_comp1 = mdl.get_item("GPIO DI (Generic)1", parent=subsys_handle)
    gpio_comp2 = mdl.get_item("GPIO DI (Generic)2", parent=subsys_handle)
    epwm_comp = mdl.get_item("ePWM (Generic)1", parent=subsys_handle)

    for comp in [adc_comp1, adc_comp2, adc_comp3, gpio_comp1, gpio_comp2, epwm_comp]:
        # Just to trigger property_value_changed handler so the platform will be taken into account
        mdl.set_property_value(
            mdl.prop(comp, "interface_type"), "HIL DSP Interface"
        )
        mdl.set_property_value(
            mdl.prop(comp, "interface_type"), interface_board
        )
        mdl.set_property_value(
            mdl.prop(comp, "controller_index"), "1"
        )

    # ADC - HIL analog output select
    current_analog_outputs = get_hil_outputs['analog']
    for adc_comp, ao in zip(
        [adc_comp1, adc_comp2, adc_comp3], current_analog_outputs
    ):
        mdl.set_property_value(
            mdl.prop(adc_comp, "hil_ao_nr"), ao
        )
        mdl.set_property_value(
            mdl.prop(adc_comp, "scaling"), scaling
        )
        mdl.set_property_value(
            mdl.prop(adc_comp, "offset"), offset
        )
        mdl.set_property_value(
            mdl.prop(adc_comp, "conv_trigger"), adc_trigger
        )
        mdl.set_property_value(
            mdl.prop(adc_comp, "inten"), True
        )

    # Always available HIL DOs
    mdl.set_property_value(
        mdl.prop(gpio_comp1, "hil_do_nr"), "6"
    )
    mdl.set_property_value(
        mdl.prop(gpio_comp2, "hil_do_nr"), "9"
    )

    mdl.set_property_value(
        mdl.prop(epwm_comp, "nr_of_channels"), "3"
    )
    hil_digital_inputs_pwm = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr")
    )
    hil_digital_inputs_pwm_b = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr_b")
    )

    build_and_deploy_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(build_and_deploy_action_prop, True)
    mdl.set_property_value(build_and_deploy_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    # HIL DI number property od ePWM (Generic) displays only HIL pins that are mapped
    # to ePWM A channels. B channel connection is dependent on interface board type (see interface documentation)
    channel_a_digital_inputs = hil_digital_inputs_pwm.split("-")
    channel_b_digital_inputs = hil_digital_inputs_pwm_b.split("-")

    converter_handle = mdl.get_item("Three Phase Inverter1", parent=None)

    # Switches that are controlled by ePWM A channels
    switches_a = ["Sa_top", "Sb_top", "Sc_top"]
    for switch, hil_di in zip(switches_a, channel_a_digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), hil_di
        )

    # Switches that are controlled by ePWM B channels
    switches_b = ["Sa_bot", "Sb_bot", "Sc_bot"]
    for switch, hil_di in zip(switches_b, channel_b_digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), hil_di
        )

    # Set phase current and encoder signals to appropriate analog and digital outputs
    output_settings = mdl.get_item("Output Settings1", parent=None)

    analog_dict_list = [
        {
            "ao_channel": ao,
            "signal": current,
            "scaling": scaling,
            "offset": offset,
            "enable_limit": "True",
            "lower_limit": ti_hw_params.ao_limits[0],
            "upper_limit": ti_hw_params.ao_limits[1],
        }
        for ao, current in zip(current_analog_outputs, ['Ia', 'Ib', 'Ic'])
    ]
    mdl.set_property_value(
        mdl.prop(output_settings, "analog_outputs"), analog_dict_list
    )

    encoder_digital_outputs = get_hil_outputs['digital']
    encoder_signals = [
        f"3 ph PMSM JMAG1.Machine Wrapper1.machine_encoder_{channel}"
        for channel in ["A", "B", "Z"]
    ]

    state_control_digital_outputs = ["6", "9"]
    state_control_signals = ["sim_start", "speed_mode"]

    digital_outputs = encoder_digital_outputs + state_control_digital_outputs
    digital_output_signals = encoder_signals + state_control_signals
    digital_dict_list = [
        {
            "do_channel": do,
            "signal": signal,
            "invert": False,
            "sw_control": False,
            "sw_value": 0
        }
        for do, signal in zip(digital_outputs, digital_output_signals)
    ]
    mdl.set_property_value(
        mdl.prop(output_settings, "digital_outputs"), digital_dict_list
    )

    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)

    serial_con = pytest.ti_serial()
    yield serial_con
    hil.stop_simulation()
    serial_con.close_connection()
    mdl.close_model()


def test_foc_pmsm(setup):
    """
    Checks if machine can achieve desired speed and torque references.
    References are provided with serial communication.
    """
    serial_con = setup

    with group_context("Speed mode"):
        # Speed mode - set HIL DO 9
        hil.set_digital_output(9, invert=False, swControl=True, value=1, device=0)

        hil.start_simulation()

        # Set speed reference
        for n in [-2e3, 5e2, 5e3, 9e3, 5e3, 5e2, -2e3]:
            with group_context(f"n = {n} rpm"):
                # Set speed reference
                serial_con.transmit(n, 1)
                hil.wait_sec(10)

                # Capturing data
                capture.start_capture(duration=2, signals=["n_meas"], rate=50e3)
                cap_data = capture.get_capture_results(wait_capture=True)

                # Data analysis
                assert_is_constant(
                    cap_data["n_meas"],
                    at_value=ranges.around(n, tol_p=0.02),
                    during=ranges.after(1e-3),
                    strictness=0.99
                )

        hil.stop_simulation()

    with group_context("Torque mode"):
        # Torque mode - clear HIL DO 9
        hil.set_digital_output(9, invert=False, swControl=True, value=0, device=0)

        for t in [150, 100, 50, -50, -100]:
            with group_context(f"T_electrical = {t} Nm"):
                # Set torque reference
                serial_con.transmit(t, 2)

                hil.start_simulation()
                hil.wait_sec(1)

                # Capturing data
                capture.start_capture(duration=0.2, signals=["electrical_torque"], rate=50000)
                cap_data = capture.get_capture_results(wait_capture=True)

                hil.stop_simulation()

                # Data analysis
                assert_is_constant(
                    cap_data["electrical_torque"],
                    at_value=ranges.around(t, tol_p=0.15),
                    during=ranges.after(1e-3),
                    strictness=0.99
                )

