import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
from typhoon.test.rms import window_rms
import logging
import pytest
import os
import numpy as np
import params


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

logger = logging.getLogger(__name__)

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located
TESTS_DIR = os.path.dirname(FILE_DIR_PATH)

MODEL_NAME = "open_loop_3_lvl_inverter.tse"

# .tse file paths
MODEL_PATH = os.path.join(
    os.path.dirname(TESTS_DIR), 'models', 'open_loop_3_lvl_inverter', MODEL_NAME
)

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)

# Digital input chains for Three Phase T Type Inverter switches
# Switch order on Three Phase T Type Inverter:
# "Sa_1", "Sa_2", "Sa_3", "Sa_4", "Sb1", "Sb_2", "Sb_3", "Sb_4", "Sc_1", "Sc_2", "Sc_3", "Sc_4".
DIGITAL_INPUTS_3LVL = {
    params.INTERFACE_LAUNCHPAD: [
        ['1', '3', '2', '4', '5', '9', '6', '10', '11', '13', '12', '14'],  # F2379D
        ['9', '13', '10', '14', '5', '11', '6', '12', '3', '1', '4', '2'],  # F280049C
        ['1', '3', '2', '4', '11', '13', '12', '14', '5', '9', '6', '10'],  # F280039C
    ],
    params.INTERFACE_UGRID: [
        ['1', '3', '2', '4', '5', '7', '6', '8', '9', '11', '10', '12'],    # F28379D
        ['7', '11', '8', '12', '5', '9', '6', '10', '3', '1', '4', '2'],    # F280049C
        ['1', '3', '2', '4', '9', '11', '10', '12', '5', '7', '6', '8'],    # F280039C
    ],
    params.INTERFACE_180: [
        ['1', '2', '7', '8', '3', '4', '9', '10', '5', '6', '11', '12']     # F28379D
    ],
    params.INTERFACE_100: [
        ['1', '2', '7', '8', '3', '4', '9', '10', '5', '6', '11', '12']     # F28335
    ]
}

EPWM_MODULES = [
    ['1', '2', '3', '4', '5', '6'],
    ['1', '2', '3', '4', '5', '6'],
    ['1', '2', '4', '5', '6', '7'],
]


def get_3lvl_pwm_di_chain(interface_board: str, device: str) -> dict:
    """
    Fixture that provides list of HIL digital inputs connected to TI ePWM modules 1-6.

    Returns:
        list: List of HIL DI numbers connected to ePWM A channels of modules 1-6 respectively, in form of strings.
    """
    if interface_board in [params.INTERFACE_LAUNCHPAD, params.INTERFACE_UGRID]:
        device_index = params.LAUNCHPAD_DEVICES.index(device)
    else:
        device_index = 0

    return {
        "digital_input_chain": DIGITAL_INPUTS_3LVL[interface_board][device_index],
        "module_chain": EPWM_MODULES[device_index]
    }


@pytest.fixture(scope="module")
def setup(get_hw_setup):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Six consecutive ePWM modules are configured with core ePWM components.
    Availability of modules may vary regarding selected interface and platform.
    Device clock is dependent on selected platform which affects frequency and dead time.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device
    processor = ti_hw_params.ti_processor

    epwm_3_lvl_coinfig = get_3lvl_pwm_di_chain(interface_board, processor)
    digital_inputs = epwm_3_lvl_coinfig["digital_input_chain"]
    epwm_modules = epwm_3_lvl_coinfig["module_chain"]

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type='component')
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    subsys_handle = mdl.get_item('open_loop', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    for i, module_nr in enumerate(epwm_modules):
        epwm_comp = mdl.get_item(f"ePWM{i + 1}", parent=subsys_handle, item_type=ITEM_COMPONENT)
        mdl.set_property_value(
            mdl.prop(epwm_comp, "epwm_module"), module_nr
        )

    button_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(button_action_prop, True)
    mdl.set_property_value(button_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    converter_handle = mdl.get_item("Three Phase T Type Inverter", parent=None)
    switches = [
        "Sa_1", "Sa_2", "Sa_3", "Sa_4",
        "Sb_1", "Sb_2", "Sb_3", "Sb_4",
        "Sc_1", "Sc_2", "Sc_3", "Sc_4"
    ]
    for switch, di_nr in zip(switches, digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), di_nr
        )

    hw_settings = mdl.detect_hw_settings()
    logger.info("{} {} device is used".format(hw_settings[0], hw_settings[2]))
    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)
    hil.start_simulation()

    serial_con = pytest.ti_serial()
    yield serial_con
    hil.stop_simulation()
    serial_con.close_connection()
    mdl.close_model()


@pytest.mark.parametrize("mod", [0.3, 0.5, 0.8, 1])
@pytest.mark.parametrize("freq", [50, 60])
def test_open_loop_3_lvl_inverter(setup, mod, freq):
    """
    Checks if input parameters of modulating signal yield expected output voltages.
    """
    serial_con = setup
    capture_signals = ["Ia"]

    # Set speed reference
    serial_con.transmit(mod, 1)
    serial_con.transmit(freq, 2)

    hil.wait_sec(1)

    # Capturing data
    capture.start_capture(
        duration=4/freq, signals=capture_signals, rate=50000
    )
    cap_data = capture.get_capture_results(wait_capture=True)
    
    i_rms = np.mean(
        window_rms(
            cap_data["Ia"], max_frequency=2*freq, match_num_of_cycles_with_samples=False, assume_previous_data=True
        )
    )

    # Data analysis
    vdc = 800
    z = complex(1, 2 * np.pi * freq * 1e-3)
    i_rms_ref = mod * vdc / np.sqrt(6) / abs(z)
    assert i_rms == pytest.approx(i_rms_ref, rel=0.05)
