import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
import typhoon.test.signals as sig
from typhoon.test.ranges import around, after
import logging
import pytest
import os


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located

tests_dir = os.path.dirname(FILE_DIR_PATH)

logger = logging.getLogger(__name__)

MODEL_NAME = "open_loop_buck.tse"

# .tse file paths
MODEL_PATH = os.path.join(
    os.path.dirname(tests_dir), 'models', 'open_loop_buck', 'open_loop_buck.tse'
)

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)


@pytest.fixture(scope="module")
def setup_function(get_hw_setup):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Generic ePWM is used to configure single pwm channel.
    Target HIL DI may vary regarding selected interface and platform.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type='component')
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    subsys_handle = mdl.get_item('open_loop', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    # Configure generic epwm comp regarding interface and platform setup
    epwm_comp = mdl.get_item("ePWM - HIL DI1", parent=subsys_handle)

    # Just to trigger property_value_changed handler so the platform will be taken into account
    mdl.set_property_value(
        mdl.prop(epwm_comp, "interface_type"), "HIL DSP Interface"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "interface_type"), interface_board
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "controller_index"), "1"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "nr_of_channels"), "1"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "hil_di_nr"), "1"
    )

    button_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(button_action_prop, True)
    mdl.set_property_value(button_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    hw_settings = mdl.detect_hw_settings()
    logger.info("{} c{} device is used".format(hw_settings[0], hw_settings[2]))
    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)

    serial_con = pytest.ti_serial()
    yield serial_con
    hil.stop_simulation()
    serial_con.close_connection()
    mdl.close_model()


@pytest.mark.parametrize("duty_cycle", [0.25, 0.5, 0.75, 0.95])
def test_open_loop_buck(setup_function, duty_cycle):
    """
    Test checks whether the output voltage of buck converter corresponds to duty cycle sent by serial communication.
    """

    serial_con = setup_function
    capture_duration = 0.1
    capture_signals = ["Vout"]
    v_in = 100

    # start simulation, wait for capture to finish, stop simulation
    hil.start_simulation()

    # Set duty cycle
    serial_con.transmit(duty_cycle, 1)

    hil.wait_msec(100)
    capture.start_capture(
        duration=capture_duration, signals=capture_signals, trigger_source="Forced", rate=500e3, timeout=10
    )

    cap_data = capture.get_capture_results(wait_capture=True)
    hil.stop_simulation()

    assert sig.is_constant(
        cap_data["Vout"], at_value=around(duty_cycle * v_in, tol_p=0.05), during=after(10e-3), strictness=0.99
    )
