import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
from typhoon.test.signals import assert_is_constant
import typhoon.test.ranges as ranges
import logging
import pytest
import os
import params


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

logger = logging.getLogger(__name__)

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located
TESTS_DIR = os.path.dirname(FILE_DIR_PATH)

MODEL_NAME = "open_loop_llc_converter.tse"

# .tse file paths
MODEL_PATH = os.path.join(
    os.path.dirname(TESTS_DIR), 'models', 'open_loop_llc_converter', MODEL_NAME
)

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)


REF_DATA = {
    0.5: 1.5,
    0.7: 1.16,
    1: 1,
    1.3: 0.88,
    1.6: 0.78,
    2: 0.69,
    2.5: 0.6
}


@pytest.fixture(scope="module")
def setup(request, get_hw_setup):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Generic ePWM is used to configure two pwm modules with 2 channels with variable frequency.
    Target HIL DIs may vary regarding selected interface and platform.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    hw_settings = mdl.detect_hw_settings()
    product = hw_settings[0]
    configuration = hw_settings[2]
    logger.info("{} c{} device is used".format(product, configuration))
    pytest.check_hw_requirements(request, product, configuration)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type='component')
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    subsys_handle = mdl.get_item('freq_mod', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    epwm_comp = mdl.get_item("ePWM (Generic)1", parent=subsys_handle, item_type=ITEM_COMPONENT)

    # Just to trigger property_value_changed handler so the platform will be taken into account
    mdl.set_property_value(
        mdl.prop(epwm_comp, "interface_type"), "HIL DSP Interface"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "interface_type"), interface_board
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "controller_index"), "1"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "nr_of_channels"), "2"
    )

    hil_digital_inputs_pwm = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr")
    )
    hil_digital_inputs_pwm_b = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr_b")
    )

    button_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(button_action_prop, True)
    mdl.set_property_value(button_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    # HIL DI number property od ePWM (Generic) displays only HIL pins that are mapped
    # to ePWM A channels. B channel connection is dependent on interface board type (see interface documentation)
    channel_a_digital_inputs = hil_digital_inputs_pwm.split("-")
    channel_b_digital_inputs = hil_digital_inputs_pwm_b.split("-")

    converter_handle = mdl.get_item("Resonant Converter1", parent=None)

    switches_a = ["SA_1", "SA_3"]
    for switch, hil_di in zip(switches_a, channel_a_digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), hil_di
        )

    # Switches that are controlled by ePWM B channels
    switches_b = ["SA_2", "SA_4"]
    for switch, hil_di in zip(switches_b, channel_b_digital_inputs):
        mdl.set_property_value(
            mdl.prop(converter_handle, switch), hil_di
        )

    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)

    serial_con = pytest.ti_serial()
    yield serial_con
    hil.stop_simulation()
    serial_con.close_connection()
    mdl.close_model()


@pytest.mark.parametrize("freq", [0.5, 0.7, 1, 1.3, 1.6, 2, 2.5])
def test_open_loop_llc_converter(setup, freq):
    """
    Checks if input parameters of modulating signal yield expected output voltages.
    """
    serial_con = setup
    sigs_to_cap = ['voltage_gain']
    cap_duration = 0.1

    # Set speed reference
    serial_con.transmit(freq, 1)
    hil.wait_sec(1)

    hil.start_simulation()

    # Capturing data
    capture.start_capture(
        duration=cap_duration, signals=sigs_to_cap, rate=50000
    )
    cap_data = capture.get_capture_results(wait_capture=True)

    hil.stop_simulation()

    with pytest.assume:
        ref_value = REF_DATA[freq]
        assert_is_constant(
            cap_data['voltage_gain'], at_value=ranges.around(ref_value, tol_p=0.05), strictness=0.99,
        )

