import typhoon.api.hil as hil
from typhoon.api.schematic_editor import model as mdl
from typhoon.api.schematic_editor.const import ITEM_COMPONENT
import typhoon.test.capture as capture
import logging
import pytest
import os
from typhoon.test.harmonic import frequency_content
import params


# IMPORTANT NOTE: Instructions on how to run the tests are located in
# examples.pdf file in documentation folder!

FILE_DIR_PATH = os.path.dirname(os.path.abspath(__file__))  # path of the directory where this script is located

tests_dir = os.path.dirname(FILE_DIR_PATH)

logger = logging.getLogger(__name__)

MODEL_NAME = "sinusoidal_pwm.tse"

# .tse file paths
MODEL_PATH = os.path.join(
    os.path.dirname(tests_dir), 'models', 'sinusoidal_pwm', 'sinusoidal_pwm.tse'
)

# .cpd file paths
COMPILED_MODEL_PATH = mdl.get_compiled_model_file(MODEL_PATH)


@pytest.fixture(scope="module")
def setup_function(get_hw_setup):
    """
    Function that loads the model, builds and deploys the code and compiles the model.

    Generic ePWM is used to configure single pwm module with 2 channels.
    Target HIL DIs may vary regarding selected interface and platform.
    """
    ti_hw_params = get_hw_setup
    interface_board = ti_hw_params.interface_name
    platform = ti_hw_params.ti_device

    mdl.reload_libraries()
    mdl.load(MODEL_PATH)

    setup_handle = mdl.get_item('TI C2000 Setup1', item_type='component')
    mdl.set_property_value(
        mdl.prop(setup_handle, "target_platform"), platform
    )

    subsys_handle = mdl.get_item('sine_pwm', item_type=ITEM_COMPONENT)
    mdl.enable_items([subsys_handle])

    epwm_comp = mdl.get_item("ePWM - HIL DI1", parent=subsys_handle)

    # Just to trigger property_value_changed handler so the platform will be taken into account
    mdl.set_property_value(
        mdl.prop(epwm_comp, "interface_type"), "HIL DSP Interface"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "interface_type"), interface_board
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "controller_index"), "1"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "nr_of_channels"), "1"
    )
    mdl.set_property_value(
        mdl.prop(epwm_comp, "hil_di_nr"), "3"
    )
    channel_b_nr = mdl.get_property_value(
        mdl.prop(epwm_comp, "hil_di_nr_b")
    )

    button_action_prop = mdl.prop(setup_handle, 'build_and_deploy_action')
    mdl.set_property_value(button_action_prop, True)
    mdl.set_property_value(button_action_prop, False)

    # Disable the subsystem that's being exported.
    mdl.disable_items([subsys_handle])

    converter_handle = mdl.get_item("IGBT Leg1", parent=None)
    mdl.set_property_value(
        mdl.prop(converter_handle, "S_top"), "3"
    )
    mdl.set_property_value(
        mdl.prop(converter_handle, "S_bot"), channel_b_nr
    )

    hw_settings = mdl.detect_hw_settings()
    mdl.compile()
    hil.load_model(COMPILED_MODEL_PATH, vhil_device=False)
    logger.info("{} c{} device is used".format(hw_settings[0], hw_settings[2]))

    yield
    hil.stop_simulation()
    mdl.close_model()


def test_sin_pwm(setup_function):
    """
    Test checks whether the output of converter corresponds to sinusoidal modulating signal.
    """

    capture_duration = 0.1
    capture_signals = ["L1"]

    # start simulation, wait for capture to finish, stop simulation
    hil.start_simulation()

    hil.wait_msec(100)
    capture.start_capture(
        duration=capture_duration, signals=capture_signals, trigger_source="L1",
        trigger_threshold=0.0, trigger_edge="rising", rate=500e3, timeout=10
    )

    cap_data = capture.get_capture_results(wait_capture=True)
    hil.stop_simulation()

    freq_cont = frequency_content(signal=cap_data["L1"], max_frequency=int(15e3))
    
    amp_fmod = abs(freq_cont(50, tol=1))
    amp_fsw = abs(freq_cont(1e4, tol=1))
    
    assert amp_fmod > 9 and amp_fsw > 0.08
