import pytest
import serial
import serial.tools.list_ports as port_list
import struct
import time
import os
import params
import typhoon.api.hil as hil


PING = b'\x05\x05\x05\x05\x05'


def check_hw_requirements(request, hil_device: str, config: int):
    test_file_name = os.path.basename(request.fspath)
    required_features = params.HW_REQUIREMENTS[test_file_name]
    if required_features:
        if not hil._check_requirements(
                required_features, device=hil_device, conf_id=config
        ):
            feature = "non-linear machines" if "pmsm" in test_file_name else "DC-DC converter solver"
            pytest.skip(
                f"{hil_device} configuration {config} does not support {feature}! "
                f"Check device table and select suitable configuration."
            )


def get_ti_sci_port() -> str:
    """
    Function that detects COM port suitable for data exchange between PC and TI device via SCI.
    """
    target_port = ""
    try:
        ports = list(port_list.comports())
        ti_ports = list(filter(lambda p: 'Texas Instruments' in p.manufacturer, ports))
        if len(ti_ports) > 1:
            for port in ti_ports:
                if 'UART' in port.description:
                    return port.name

            return sorted(ti_ports, key=lambda p: int(p.name[3:]))[-1].name

        elif len(ti_ports) == 1:
            return ti_ports[0].name

    except:
        pass

    return target_port


class TIPySerial:
    def __init__(self, baudrate: int = 460800, parity: str = 'N', stopbits: int = 1):
        self.target_com_port = get_ti_sci_port()
        if not self.target_com_port:
            pytest.skip("Texas Instruments device not detected. Check Windows Device Manager.")

        try:
            self.serial_con = serial.Serial(
                port=self.target_com_port, baudrate=baudrate, parity=parity, stopbits=stopbits, timeout=0.5
            )
            self.serial_con.reset_input_buffer()
            self.serial_con.reset_output_buffer()
        except Exception:
            pytest.skip("Serial communication with TI DSP cannot be established. Check Windows Device Manager.")

    def transmit(self, data: list | float, index: int = 1) -> None:
        """
        Read specified number of messages from COM port.

        Args:
            data (list | float): List of messages to send or single message to transmit to controller.
            index (int): Index of message in case single message is transmitted.

        Returns:
            None
        """
        frames = []
        if isinstance(data, list):
            if len(data) > 4:
                pytest.skip("Maximum amount of 4 messages exceeded!")

            for i, msg in enumerate(data):
                try:
                    msg_float = float(msg)
                except ValueError:
                    pytest.skip(
                        f"Message {msg} has invalid type. Supported types data exchange via SCI: int, float"
                    )
                msg_id = (i + 1).to_bytes(1, 'big')
                tx_data = struct.pack('f', msg_float)
                frames.append(msg_id + tx_data)

        else:
            try:
                msg = float(data)
                msg_id = int(index)
            except ValueError:
                pytest.skip(
                    "Invalid parameter type in serial transmit function. "
                    "Supported types data exchange via SCI:\n"
                    "data: int, float\n "
                    "index: int"
                )
            msg_id = (msg_id).to_bytes(1, 'big')
            tx_data = struct.pack('f', msg)
            frames.append(msg_id + tx_data)

        for frame in frames:
            try:
                bytes_written = self.serial_con.write(frame)
                time.sleep(0.01)
            except Exception:
                pytest.fail("Failed to write message.")

            if bytes_written == 0:
                pytest.fail("Failed to write message.")

    def receive(self, nr_of_msg: int, ping: bool = True) -> list:
        """
        Read specified number of messages from COM port.

        Args:
            nr_of_msg (int): Number of messages to read. Every message contains 4 bytes (32 bits).
            ping (bool): Decides whether to send ping to controller. Default value is True

        Returns:
            list: List of decoded messages.
        """
        received_data = []
        if nr_of_msg > 4:
            pytest.fail("Maximum amount of 4 messages exceeded!")

        if ping:
            self.serial_con.write(PING)
            time.sleep(0.5)

        try:
            # read data from com port (4 bytes per vector element)
            rx_data = self.serial_con.read(nr_of_msg * 4)
        except Exception:
            pytest.fail("Failed to read message.")

        if len(rx_data) < nr_of_msg * 4:
            pytest.fail(f"Data lost while receiving. Sent {nr_of_msg * 4} bytes and received {len(rx_data)} bytes.")

        for i in range(nr_of_msg):
            unpacked_data = struct.unpack('f', rx_data[i * 4:i * 4 + 4])[0]
            received_data.append(round(unpacked_data, 4))

        return received_data

    def close_connection(self):
        """
        Clear input and output buffers and closes the connection.
        """
        self.serial_con.reset_input_buffer()
        self.serial_con.reset_output_buffer()
        self.serial_con.close()